@@ -98,7 +98,7 @@
 #define YYMALLOC palloc
 #define YYFREE   pfree
 
-static float seg_atof(char *value);
+static float seg_atof(const char *value);
 
 static char strbuf[25] = {
 	'0', '0', '0', '0', '0',
@@ -162,16 +162,17 @@ extern int seg_yydebug;
 #if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
 union YYSTYPE
 {
-#line 41 "segparse.y"
+#line 42 "segparse.y"
 
-	struct BND {
-		float val;
-		char  ext;
-		char  sigd;
+	struct BND
+	{
+		float		val;
+		char		ext;
+		char		sigd;
 	} bnd;
-	char * text;
+	char	   *text;
 
-#line 175 "segparse.c"
+#line 176 "segparse.c"
 
 };
 typedef union YYSTYPE YYSTYPE;
@@ -580,7 +581,7 @@ static const yytype_int8 yytranslate[] =
   /* YYRLINE[YYN] -- Source line where rule number YYN was defined.  */
 static const yytype_uint8 yyrline[] =
 {
-       0,    61,    61,    73,    91,   101,   111,   119,   128,   139
+       0,    63,    63,    75,    93,   103,   113,   121,   130,   141
 };
 #endif
 
@@ -1157,7 +1158,7 @@ yyreduce:
   switch (yyn)
     {
   case 2: /* range: boundary PLUMIN deviation  */
-#line 62 "segparse.y"
+#line 64 "segparse.y"
         {
 		result->lower = (yyvsp[-2].bnd).val - (yyvsp[0].bnd).val;
 		result->upper = (yyvsp[-2].bnd).val + (yyvsp[0].bnd).val;
@@ -1168,11 +1169,11 @@ yyreduce:
 		result->l_ext = '\0';
 		result->u_ext = '\0';
 	}
-#line 1172 "segparse.c"
+#line 1173 "segparse.c"
     break;
 
   case 3: /* range: boundary RANGE boundary  */
-#line 74 "segparse.y"
+#line 76 "segparse.y"
         {
 		result->lower = (yyvsp[-2].bnd).val;
 		result->upper = (yyvsp[0].bnd).val;
@@ -1189,11 +1190,11 @@ yyreduce:
 		result->l_ext = ( (yyvsp[-2].bnd).ext ? (yyvsp[-2].bnd).ext : '\0' );
 		result->u_ext = ( (yyvsp[0].bnd).ext ? (yyvsp[0].bnd).ext : '\0' );
 	}
-#line 1193 "segparse.c"
+#line 1194 "segparse.c"
     break;
 
   case 4: /* range: boundary RANGE  */
-#line 92 "segparse.y"
+#line 94 "segparse.y"
         {
 		result->lower = (yyvsp[-1].bnd).val;
 		result->upper = HUGE_VAL;
@@ -1202,11 +1203,11 @@ yyreduce:
 		result->l_ext = ( (yyvsp[-1].bnd).ext ? (yyvsp[-1].bnd).ext : '\0' );
 		result->u_ext = '-';
 	}
-#line 1206 "segparse.c"
+#line 1207 "segparse.c"
     break;
 
   case 5: /* range: RANGE boundary  */
-#line 102 "segparse.y"
+#line 104 "segparse.y"
         {
 		result->lower = -HUGE_VAL;
 		result->upper = (yyvsp[0].bnd).val;
@@ -1215,60 +1216,60 @@ yyreduce:
 		result->l_ext = '-';
 		result->u_ext = ( (yyvsp[0].bnd).ext ? (yyvsp[0].bnd).ext : '\0' );
 	}
-#line 1219 "segparse.c"
+#line 1220 "segparse.c"
     break;
 
   case 6: /* range: boundary  */
-#line 112 "segparse.y"
+#line 114 "segparse.y"
         {
 		result->lower = result->upper = (yyvsp[0].bnd).val;
 		result->l_sigd = result->u_sigd = (yyvsp[0].bnd).sigd;
 		result->l_ext = result->u_ext = ( (yyvsp[0].bnd).ext ? (yyvsp[0].bnd).ext : '\0' );
 	}
-#line 1229 "segparse.c"
+#line 1230 "segparse.c"
     break;
 
   case 7: /* boundary: SEGFLOAT  */
-#line 120 "segparse.y"
+#line 122 "segparse.y"
         {
 		/* temp variable avoids a gcc 3.3.x bug on Sparc64 */
-		float val = seg_atof((yyvsp[0].text));
+		float		val = seg_atof((yyvsp[0].text));
 
 		(yyval.bnd).ext = '\0';
 		(yyval.bnd).sigd = significant_digits((yyvsp[0].text));
 		(yyval.bnd).val = val;
 	}
-#line 1242 "segparse.c"
+#line 1243 "segparse.c"
     break;
 
   case 8: /* boundary: EXTENSION SEGFLOAT  */
-#line 129 "segparse.y"
+#line 131 "segparse.y"
         {
 		/* temp variable avoids a gcc 3.3.x bug on Sparc64 */
-		float val = seg_atof((yyvsp[0].text));
+		float		val = seg_atof((yyvsp[0].text));
 
 		(yyval.bnd).ext = (yyvsp[-1].text)[0];
 		(yyval.bnd).sigd = significant_digits((yyvsp[0].text));
 		(yyval.bnd).val = val;
 	}
-#line 1255 "segparse.c"
+#line 1256 "segparse.c"
     break;
 
   case 9: /* deviation: SEGFLOAT  */
-#line 140 "segparse.y"
+#line 142 "segparse.y"
         {
 		/* temp variable avoids a gcc 3.3.x bug on Sparc64 */
-		float val = seg_atof((yyvsp[0].text));
+		float		val = seg_atof((yyvsp[0].text));
 
 		(yyval.bnd).ext = '\0';
 		(yyval.bnd).sigd = significant_digits((yyvsp[0].text));
 		(yyval.bnd).val = val;
 	}
-#line 1268 "segparse.c"
+#line 1269 "segparse.c"
     break;
 
 
-#line 1272 "segparse.c"
+#line 1273 "segparse.c"
 
       default: break;
     }
@@ -1462,14 +1463,14 @@ yyreturn:
   return yyresult;
 }
 
-#line 150 "segparse.y"
+#line 152 "segparse.y"
 
 
 
 static float
-seg_atof(char *value)
+seg_atof(const char *value)
 {
-	Datum datum;
+	Datum		datum;
 
 	datum = DirectFunctionCall1(float4in, CStringGetDatum(value));
 	return DatumGetFloat4(datum);
