@@ -3,7 +3,7 @@
  * execUtils.c
  *	  miscellaneous executor utility routines
  *
- * Portions Copyright (c) 1996-2021, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2022, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
@@ -52,7 +52,6 @@
 #include "access/transam.h"
 #include "executor/executor.h"
 #include "executor/execPartition.h"
-#include "executor/nodeModifyTable.h"
 #include "jit/jit.h"
 #include "mb/pg_wchar.h"
 #include "miscadmin.h"
@@ -66,7 +65,7 @@
 #include "utils/typcache.h"
 
 
-static bool tlist_matches_tupdesc(PlanState *ps, List *tlist, Index varno, TupleDesc tupdesc);
+static bool tlist_matches_tupdesc(PlanState *ps, List *tlist, int varno, TupleDesc tupdesc);
 static void ShutdownExprContext(ExprContext *econtext, bool isCommit);
 
 
@@ -130,11 +129,6 @@ CreateExecutorState(void)
 	estate->es_tuple_routing_result_relations = NIL;
 	estate->es_trig_target_relations = NIL;
 
-	estate->es_insert_pending_result_relations = NIL;
-	estate->es_insert_pending_modifytables = NIL;
-
-	estate->es_resultrelinfo_extra = NIL;
-
 	estate->es_param_list_info = NULL;
 	estate->es_param_exec_vals = NULL;
 
@@ -559,7 +553,7 @@ ExecAssignProjectionInfo(PlanState *planstate,
  */
 void
 ExecConditionalAssignProjectionInfo(PlanState *planstate, TupleDesc inputDesc,
-									Index varno)
+									int varno)
 {
 	if (tlist_matches_tupdesc(planstate,
 							  planstate->plan->targetlist,
@@ -585,7 +579,7 @@ ExecConditionalAssignProjectionInfo(PlanState *planstate, TupleDesc inputDesc,
 }
 
 static bool
-tlist_matches_tupdesc(PlanState *ps, List *tlist, Index varno, TupleDesc tupdesc)
+tlist_matches_tupdesc(PlanState *ps, List *tlist, int varno, TupleDesc tupdesc)
 {
 	int			numattrs = tupdesc->natts;
 	int			attrno;
@@ -782,7 +776,7 @@ ExecInitRangeTable(EState *estate, List *rangeTable)
  * ExecGetRangeTableRelation
  *		Open the Relation for a range table entry, if not already done
  *
- * The Relations will be closed in ExecEndPlan().
+ * The Relations will be closed again in ExecEndPlan().
  */
 Relation
 ExecGetRangeTableRelation(EState *estate, Index rti)
@@ -1326,47 +1320,32 @@ ExecGetUpdatedCols(ResultRelInfo *relinfo, EState *estate)
 Bitmapset *
 ExecGetExtraUpdatedCols(ResultRelInfo *relinfo, EState *estate)
 {
-	Relation	rel = relinfo->ri_RelationDesc;
-	TupleDesc	tupdesc = RelationGetDescr(rel);
-
-	if (tupdesc->constr && tupdesc->constr->has_generated_stored)
+	/* see ExecGetInsertedCols() */
+	if (relinfo->ri_RangeTableIndex != 0)
 	{
-		ListCell   *lc;
+		RangeTblEntry *rte = exec_rt_fetch(relinfo->ri_RangeTableIndex, estate);
 
-		/* Compute the info if we didn't already */
-		if (relinfo->ri_GeneratedExprs == NULL)
-			ExecInitStoredGenerated(relinfo, estate, CMD_UPDATE);
-		foreach(lc, estate->es_resultrelinfo_extra)
-		{
-			ResultRelInfoExtra *rextra = (ResultRelInfoExtra *) lfirst(lc);
-
-			if (rextra->rinfo == relinfo)
-				return rextra->ri_extraUpdatedCols;
-		}
-		Assert(false);			/* shouldn't get here */
+		return rte->extraUpdatedCols;
 	}
-	return NULL;
+	else if (relinfo->ri_RootResultRelInfo)
+	{
+		ResultRelInfo *rootRelInfo = relinfo->ri_RootResultRelInfo;
+		RangeTblEntry *rte = exec_rt_fetch(rootRelInfo->ri_RangeTableIndex, estate);
+
+		if (relinfo->ri_RootToPartitionMap != NULL)
+			return execute_attr_map_cols(relinfo->ri_RootToPartitionMap->attrMap,
+										 rte->extraUpdatedCols);
+		else
+			return rte->extraUpdatedCols;
+	}
+	else
+		return NULL;
 }
 
-/*
- * Return columns being updated, including generated columns
- *
- * The bitmap is allocated in per-tuple memory context. It's up to the caller to
- * copy it into a different context with the appropriate lifespan, if needed.
- */
+/* Return columns being updated, including generated columns */
 Bitmapset *
 ExecGetAllUpdatedCols(ResultRelInfo *relinfo, EState *estate)
 {
-
-	Bitmapset	   *ret;
-	MemoryContext	oldcxt;
-
-	oldcxt = MemoryContextSwitchTo(GetPerTupleMemoryContext(estate));
-
-	ret = bms_union(ExecGetUpdatedCols(relinfo, estate),
-					ExecGetExtraUpdatedCols(relinfo, estate));
-
-	MemoryContextSwitchTo(oldcxt);
-
-	return ret;
+	return bms_union(ExecGetUpdatedCols(relinfo, estate),
+					 ExecGetExtraUpdatedCols(relinfo, estate));
 }
