@@ -107,10 +107,10 @@
  * is a convenient point to initialize replication from, which is why we
  * export a snapshot at that point, which *can* be used to read normal data.
  *
- * Copyright (c) 2012-2021, PostgreSQL Global Development Group
+ * Copyright (c) 2012-2022, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
- *	  src/backend/replication/snapbuild.c
+ *	  src/backend/replication/logical/snapbuild.c
  *
  *-------------------------------------------------------------------------
  */
@@ -165,15 +165,15 @@ struct SnapBuild
 	XLogRecPtr	start_decoding_at;
 
 	/*
-	 * LSN at which we found a consistent point at the time of slot creation.
-	 * This is also the point where we have exported a snapshot for the
-	 * initial copy.
+	 * LSN at which two-phase decoding was enabled or LSN at which we found a
+	 * consistent point at the time of slot creation.
 	 *
-	 * The prepared transactions that are not covered by initial snapshot
-	 * needs to be sent later along with commit prepared and they must be
-	 * before this point.
+	 * The prepared transactions, that were skipped because previously
+	 * two-phase was not enabled or are not covered by initial snapshot, need
+	 * to be sent later along with commit prepared and they must be before
+	 * this point.
 	 */
-	XLogRecPtr	initial_consistent_point;
+	XLogRecPtr	two_phase_at;
 
 	/*
 	 * Don't start decoding WAL until the "xl_running_xacts" information
@@ -290,7 +290,7 @@ static void SnapBuildFreeSnapshot(Snapshot snap);
 
 static void SnapBuildSnapIncRefcount(Snapshot snap);
 
-static void SnapBuildDistributeSnapshotAndInval(SnapBuild *builder, XLogRecPtr lsn, TransactionId xid);
+static void SnapBuildDistributeNewCatalogSnapshot(SnapBuild *builder, XLogRecPtr lsn);
 
 /* xlog reading helper functions for SnapBuildProcessRunningXacts */
 static bool SnapBuildFindSnapshot(SnapBuild *builder, XLogRecPtr lsn, xl_running_xacts *running);
@@ -300,17 +300,6 @@ static void SnapBuildWaitSnapshot(xl_running_xacts *running, TransactionId cutof
 static void SnapBuildSerialize(SnapBuild *builder, XLogRecPtr lsn);
 static bool SnapBuildRestore(SnapBuild *builder, XLogRecPtr lsn);
 
-/*
- * Memory context reset callback for clearing the array of running transactions
- * and subtransactions.
- */
-static void
-SnapBuildResetRunningXactsCallback(void *arg)
-{
-	NInitialRunningXacts = 0;
-	InitialRunningXacts = NULL;
-}
-
 /*
  * Allocate a new snapshot builder.
  *
@@ -322,12 +311,11 @@ AllocateSnapshotBuilder(ReorderBuffer *reorder,
 						TransactionId xmin_horizon,
 						XLogRecPtr start_lsn,
 						bool need_full_snapshot,
-						XLogRecPtr initial_consistent_point)
+						XLogRecPtr two_phase_at)
 {
 	MemoryContext context;
 	MemoryContext oldcontext;
 	SnapBuild  *builder;
-	MemoryContextCallback *mcallback;
 
 	/* allocate memory in own context, to have better accountability */
 	context = AllocSetContextCreate(CurrentMemoryContext,
@@ -351,17 +339,10 @@ AllocateSnapshotBuilder(ReorderBuffer *reorder,
 	builder->initial_xmin_horizon = xmin_horizon;
 	builder->start_decoding_at = start_lsn;
 	builder->building_full_snapshot = need_full_snapshot;
-	builder->initial_consistent_point = initial_consistent_point;
-
-	mcallback = palloc0(sizeof(MemoryContextCallback));
-	mcallback->func = SnapBuildResetRunningXactsCallback;
-	MemoryContextRegisterResetCallback(CurrentMemoryContext, mcallback);
+	builder->two_phase_at = two_phase_at;
 
 	MemoryContextSwitchTo(oldcontext);
 
-	/* The initial running transactions array must be empty. */
-	Assert(NInitialRunningXacts == 0 && InitialRunningXacts == NULL);
-
 	return builder;
 }
 
@@ -419,12 +400,21 @@ SnapBuildCurrentState(SnapBuild *builder)
 }
 
 /*
- * Return the LSN at which the snapshot was exported
+ * Return the LSN at which the two-phase decoding was first enabled.
  */
 XLogRecPtr
-SnapBuildInitialConsistentPoint(SnapBuild *builder)
+SnapBuildGetTwoPhaseAt(SnapBuild *builder)
 {
-	return builder->initial_consistent_point;
+	return builder->two_phase_at;
+}
+
+/*
+ * Set the LSN at which two-phase decoding is enabled.
+ */
+void
+SnapBuildSetTwoPhaseAt(SnapBuild *builder, XLogRecPtr ptr)
+{
+	builder->two_phase_at = ptr;
 }
 
 /*
@@ -843,15 +833,15 @@ SnapBuildProcessNewCid(SnapBuild *builder, TransactionId xid,
 }
 
 /*
- * Add a new Snapshot and invalidation messages to all transactions we're
- * decoding that currently are in-progress so they can see new catalog contents
- * made by the transaction that just committed. This is necessary because those
- * in-progress transactions will use the new catalog's contents from here on
- * (at the very least everything they do needs to be compatible with newer
- * catalog contents).
+ * Add a new Snapshot to all transactions we're decoding that currently are
+ * in-progress so they can see new catalog contents made by the transaction
+ * that just committed. This is necessary because those in-progress
+ * transactions will use the new catalog's contents from here on (at the very
+ * least everything they do needs to be compatible with newer catalog
+ * contents).
  */
 static void
-SnapBuildDistributeSnapshotAndInval(SnapBuild *builder, XLogRecPtr lsn, TransactionId xid)
+SnapBuildDistributeNewCatalogSnapshot(SnapBuild *builder, XLogRecPtr lsn)
 {
 	dlist_iter	txn_i;
 	ReorderBufferTXN *txn;
@@ -859,8 +849,7 @@ SnapBuildDistributeSnapshotAndInval(SnapBuild *builder, XLogRecPtr lsn, Transact
 	/*
 	 * Iterate through all toplevel transactions. This can include
 	 * subtransactions which we just don't yet know to be that, but that's
-	 * fine, they will just get an unnecessary snapshot and invalidations
-	 * queued.
+	 * fine, they will just get an unnecessary snapshot queued.
 	 */
 	dlist_foreach(txn_i, &builder->reorder->toplevel_by_lsn)
 	{
@@ -873,14 +862,6 @@ SnapBuildDistributeSnapshotAndInval(SnapBuild *builder, XLogRecPtr lsn, Transact
 		 * transaction which in turn implies we don't yet need a snapshot at
 		 * all. We'll add a snapshot when the first change gets queued.
 		 *
-		 * Similarly, we don't need to add invalidations to a transaction whose
-		 * base snapshot is not yet set. Once a base snapshot is built, it will
-		 * include the xids of committed transactions that have modified the
-		 * catalog, thus reflecting the new catalog contents. The existing
-		 * catalog cache will have already been invalidated after processing
-		 * the invalidations in the transaction that modified catalogs,
-		 * ensuring that a fresh cache is constructed during decoding.
-		 *
 		 * NB: This works correctly even for subtransactions because
 		 * ReorderBufferAssignChild() takes care to transfer the base snapshot
 		 * to the top-level transaction, and while iterating the changequeue
@@ -890,13 +871,13 @@ SnapBuildDistributeSnapshotAndInval(SnapBuild *builder, XLogRecPtr lsn, Transact
 			continue;
 
 		/*
-		 * We don't need to add snapshot or invalidations to prepared
-		 * transactions as they should not see the new catalog contents.
+		 * We don't need to add snapshot to prepared transactions as they
+		 * should not see the new catalog contents.
 		 */
 		if (rbtxn_prepared(txn) || rbtxn_skip_prepared(txn))
 			continue;
 
-		elog(DEBUG2, "adding a new snapshot and invalidations to %u at %X/%X",
+		elog(DEBUG2, "adding a new snapshot to %u at %X/%X",
 			 txn->xid, LSN_FORMAT_ARGS(lsn));
 
 		/*
@@ -906,41 +887,6 @@ SnapBuildDistributeSnapshotAndInval(SnapBuild *builder, XLogRecPtr lsn, Transact
 		SnapBuildSnapIncRefcount(builder->snapshot);
 		ReorderBufferAddSnapshot(builder->reorder, txn->xid, lsn,
 								 builder->snapshot);
-
-		/*
-		 * Add invalidation messages to the reorder buffer of in-progress
-		 * transactions except the current committed transaction, for which we
-		 * will execute invalidations at the end.
-		 *
-		 * It is required, otherwise, we will end up using the stale catcache
-		 * contents built by the current transaction even after its decoding,
-		 * which should have been invalidated due to concurrent catalog
-		 * changing transaction.
-		 *
-		 * Distribute only the invalidation messages generated by the current
-		 * committed transaction. Invalidation messages received from other
-		 * transactions would have already been propagated to the relevant
-		 * in-progress transactions. This transaction would have processed
-		 * those invalidations, ensuring that subsequent transactions observe
-		 * a consistent cache state.
-		 */
-		if (txn->xid != xid)
-		{
-			uint32 ninvalidations;
-			SharedInvalidationMessage *msgs = NULL;
-
-			ninvalidations = ReorderBufferGetInvalidations(builder->reorder,
-														   xid, &msgs);
-
-			if (ninvalidations > 0)
-			{
-				Assert(msgs != NULL);
-
-				ReorderBufferAddDistributedInvalidations(builder->reorder,
-														 txn->xid, lsn,
-														 ninvalidations, msgs);
-			}
-		}
 	}
 }
 
@@ -1155,9 +1101,6 @@ SnapBuildCommitTxn(SnapBuild *builder, XLogRecPtr lsn, TransactionId xid,
 	else if (sub_needs_timetravel)
 	{
 		/* track toplevel txn as well, subxact alone isn't meaningful */
-		elog(DEBUG2, "forced transaction %u to do timetravel due to one of its subtransactions",
-			 xid);
-		needs_timetravel = true;
 		SnapBuildAddCommittedTxn(builder, xid);
 	}
 	else if (needs_timetravel)
@@ -1219,11 +1162,8 @@ SnapBuildCommitTxn(SnapBuild *builder, XLogRecPtr lsn, TransactionId xid,
 		/* refcount of the snapshot builder for the new snapshot */
 		SnapBuildSnapIncRefcount(builder->snapshot);
 
-		/*
-		 * Add a new catalog snapshot and invalidations messages to all
-		 * currently running transactions.
-		 */
-		SnapBuildDistributeSnapshotAndInval(builder, lsn, xid);
+		/* add a new catalog snapshot to all currently running transactions */
+		SnapBuildDistributeNewCatalogSnapshot(builder, lsn);
 	}
 }
 
@@ -1343,8 +1283,6 @@ SnapBuildProcessRunningXacts(SnapBuild *builder, XLogRecPtr lsn, xl_running_xact
 static bool
 SnapBuildFindSnapshot(SnapBuild *builder, XLogRecPtr lsn, xl_running_xacts *running)
 {
-	LogicalDecodingContext *ctx = (LogicalDecodingContext *) builder->reorder->private_data;
-
 	/* ---
 	 * Build catalog decoding snapshot incrementally using information about
 	 * the currently running transactions. There are several ways to do that:
@@ -1354,12 +1292,10 @@ SnapBuildFindSnapshot(SnapBuild *builder, XLogRecPtr lsn, xl_running_xacts *runn
 	 *	  state while waiting on c)'s sub-states.
 	 *
 	 * b) This (in a previous run) or another decoding slot serialized a
-	 *	  snapshot to disk that we can use. Can't use this method while finding
-	 *	  the start point for decoding changes as the restart LSN would be an
-	 *	  arbitrary LSN but we need to find the start point to extract changes
-	 *	  where we won't see the data for partial transactions. Also, we cannot
-	 *	  use this method when a slot needs a full snapshot for export or direct
-	 *	  use, as that snapshot will only contain catalog modifying transactions.
+	 *	  snapshot to disk that we can use.  Can't use this method for the
+	 *	  initial snapshot when slot is being created and needs full snapshot
+	 *	  for export or direct use, as that snapshot will only contain catalog
+	 *	  modifying transactions.
 	 *
 	 * c) First incrementally build a snapshot for catalog tuples
 	 *	  (BUILDING_SNAPSHOT), that requires all, already in-progress,
@@ -1424,13 +1360,8 @@ SnapBuildFindSnapshot(SnapBuild *builder, XLogRecPtr lsn, xl_running_xacts *runn
 
 		return false;
 	}
-
-	/*
-	 * b) valid on disk state and while neither building full snapshot nor
-	 * creating a slot.
-	 */
+	/* b) valid on disk state and not building full snapshot */
 	else if (!builder->building_full_snapshot &&
-			 !ctx->in_create &&
 			 SnapBuildRestore(builder, lsn))
 	{
 		int			nxacts = running->subxcnt + running->xcnt;
@@ -1541,7 +1472,6 @@ SnapBuildFindSnapshot(SnapBuild *builder, XLogRecPtr lsn, xl_running_xacts *runn
 	 * records so incremental cleanup can be performed.
 	 */
 	return true;
-
 }
 
 /* ---
@@ -1599,7 +1529,6 @@ SnapBuildWaitSnapshot(xl_running_xacts *running, TransactionId cutoff)
  * We store current state of struct SnapBuild on disk in the following manner:
  *
  * struct SnapBuildOnDisk;
- * TransactionId * running.xcnt_space;
  * TransactionId * committed.xcnt; (*not xcnt_space*)
  *
  */
@@ -2029,12 +1958,8 @@ SnapBuildRestore(SnapBuild *builder, XLogRecPtr lsn)
 	if (TransactionIdPrecedes(ondisk.builder.xmin, builder->initial_xmin_horizon))
 		goto snapshot_not_interesting;
 
-	/*
-	 * Consistent snapshots have no next phase. Reset next_phase_at as it is
-	 * possible that an old value may remain.
-	 */
+	/* consistent snapshots have no next phase */
 	Assert(ondisk.builder.next_phase_at == InvalidTransactionId);
-	builder->next_phase_at = InvalidTransactionId;
 
 	/* ok, we think the snapshot is sensible, copy over everything important */
 	builder->xmin = ondisk.builder.xmin;
@@ -2179,13 +2104,11 @@ SnapBuildXidSetCatalogChanges(SnapBuild *builder, TransactionId xid, int subxcnt
 							  TransactionId *subxacts, XLogRecPtr lsn)
 {
 	/*
-	 * Skip if there is no initial running xacts information.
-	 *
-	 * Even if the transaction has been marked as containing catalog
-	 * changes, it cannot be skipped because its subtransactions that
-	 * modified the catalog may not be marked.
+	 * Skip if there is no initial running xacts information or the
+	 * transaction is already marked as containing catalog changes.
 	 */
-	if (NInitialRunningXacts == 0)
+	if (NInitialRunningXacts == 0 ||
+		ReorderBufferXidHasCatalogChanges(builder->reorder, xid))
 		return;
 
 	if (bsearch(&xid, InitialRunningXacts, NInitialRunningXacts,
@@ -2194,6 +2117,9 @@ SnapBuildXidSetCatalogChanges(SnapBuild *builder, TransactionId xid, int subxcnt
 		ReorderBufferXidSetCatalogChanges(builder->reorder, xid, lsn);
 
 		for (int i = 0; i < subxcnt; i++)
+		{
+			ReorderBufferAssignChild(builder->reorder, xid, subxacts[i], lsn);
 			ReorderBufferXidSetCatalogChanges(builder->reorder, subxacts[i], lsn);
+		}
 	}
 }
