@@ -8,7 +8,7 @@
  * or call FUNCAPI-callable functions or macros.
  *
  *
- * Copyright (c) 2002-2019, PostgreSQL Global Development Group
+ * Copyright (c) 2002-2022, PostgreSQL Global Development Group
  *
  * src/include/funcapi.h
  *
@@ -17,11 +17,10 @@
 #ifndef FUNCAPI_H
 #define FUNCAPI_H
 
-#include "fmgr.h"
 #include "access/tupdesc.h"
 #include "executor/executor.h"
 #include "executor/tuptable.h"
-
+#include "fmgr.h"
 
 /*-------------------------------------------------------------------------
  *	Support to ease writing Functions returning composite types
@@ -279,14 +278,20 @@ extern Datum HeapTupleHeaderGetDatum(HeapTupleHeader tuple);
  * memory allocated in multi_call_memory_ctx, but holding file descriptors or
  * other non-memory resources open across calls is a bug.  SRFs that need
  * such resources should not use these macros, but instead populate a
- * tuplestore during a single call, and return that using SFRM_Materialize
- * mode (see fmgr/README).  Alternatively, set up a callback to release
- * resources at query shutdown, using RegisterExprContextCallback().
+ * tuplestore during a single call, as set up by SetSingleFuncCall() (see
+ * fmgr/README).  Alternatively, set up a callback to release resources
+ * at query shutdown, using RegisterExprContextCallback().
  *
  *----------
  */
 
 /* from funcapi.c */
+
+/* flag bits for SetSingleFuncCall() */
+#define SRF_SINGLE_USE_EXPECTED	0x01	/* use expectedDesc as tupdesc */
+#define SRF_SINGLE_BLESS		0x02	/* validate tuple for SRF */
+extern void SetSingleFuncCall(FunctionCallInfo fcinfo, bits32 flags);
+
 extern FuncCallContext *init_MultiFuncCall(PG_FUNCTION_ARGS);
 extern FuncCallContext *per_MultiFuncCall(PG_FUNCTION_ARGS);
 extern void end_MultiFuncCall(PG_FUNCTION_ARGS, FuncCallContext *funcctx);
