@@ -4,7 +4,7 @@
  *		parse analysis for utility commands
  *
  *
- * Portions Copyright (c) 1996-2018, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2022, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * src/include/parser/parse_utilcmd.h
@@ -16,23 +16,28 @@
 
 #include "parser/parse_node.h"
 
+struct AttrMap;					/* avoid including attmap.h here */
+
 
 extern List *transformCreateStmt(CreateStmt *stmt, const char *queryString);
-extern List *transformAlterTableStmt(Oid relid, AlterTableStmt *stmt,
-						const char *queryString);
+extern AlterTableStmt *transformAlterTableStmt(Oid relid, AlterTableStmt *stmt,
+											   const char *queryString,
+											   List **beforeStmts,
+											   List **afterStmts);
 extern IndexStmt *transformIndexStmt(Oid relid, IndexStmt *stmt,
-				   const char *queryString);
+									 const char *queryString);
+extern CreateStatsStmt *transformStatsStmt(Oid relid, CreateStatsStmt *stmt,
+										   const char *queryString);
 extern void transformRuleStmt(RuleStmt *stmt, const char *queryString,
-				  List **actions, Node **whereClause);
-extern List *transformCreateSchemaStmtElements(List *schemaElts,
-											   const char *schemaName);
+							  List **actions, Node **whereClause);
+extern List *transformCreateSchemaStmt(CreateSchemaStmt *stmt);
 extern PartitionBoundSpec *transformPartitionBound(ParseState *pstate, Relation parent,
-						PartitionBoundSpec *spec);
+												   PartitionBoundSpec *spec);
 extern List *expandTableLikeClause(RangeVar *heapRel,
 								   TableLikeClause *table_like_clause);
-extern IndexStmt *generateClonedIndexStmt(RangeVar *heapRel, Oid heapOid,
-						Relation source_idx,
-						const AttrNumber *attmap, int attmap_length,
-						Oid *constraintOid);
+extern IndexStmt *generateClonedIndexStmt(RangeVar *heapRel,
+										  Relation source_idx,
+										  const struct AttrMap *attmap,
+										  Oid *constraintOid);
 
 #endif							/* PARSE_UTILCMD_H */
